% This program is related to the example of turbo codes of chapter 7, the use of the logMAP / BCJR algorithm 
% Authors: Damian Gustavo Levin, Patrick Farrell and Jorge Castieira 
% 5/10/2003
% updated 03/11/2006
%--------------------------------------------------------------------------------------------
% this program is intended for academic proposes, and it is not optimised
% in terms of the programing style

clear all;
close all;
Errores=0;
Mat=zeros(4,4,3);

% Mat is an array that contains the trellis information
% the first column contains the input data, the second column contains the first output 
%(systematic bit) and the third column conains the second output
%(redundancy bit). This is the trellis of the SRC (7,5) code of figure 7.14 in chapter 7
%branch of the trellis described by the pair (current state, next state)
%Mat(current state, next state, 1)= value of the input bit that is associated to this branch 
% for the input or systematic bit

Mat(1,1,1)= -1; Mat(1,2,1)= +1;                     
Mat(2,3,1)= -1; Mat(2,4,1)= +1;
Mat(3,1,1)= +1; Mat(3,2,1)= -1;
Mat(4,3,1)= +1; Mat(4,4,1)= -1;

% first output bit of the SRC placed at column 2, Mat(current state, next
% state, 2)

Mat(1,1,2)= -1; Mat(1,2,2)= +1;                     
Mat(2,3,2)= -1; Mat(2,4,2)= +1;
Mat(3,1,2)= +1; Mat(3,2,2)= -1;
Mat(4,3,2)= +1; Mat(4,4,2)= -1;

% second output bit of the SRC placed at column 3, Mat(current state, next
% state, 3)
Mat(1,1,3)= -1; Mat(1,2,3)= +1;                       
Mat(2,3,3)= +1; Mat(2,4,3)= -1;
Mat(3,1,3)= +1; Mat(3,2,3)= -1;
Mat(4,3,3)= -1; Mat(4,4,3)= +1;

%Numerical Constants 

N=16;     % length of the block to be encoded
%Inter=[1 4 7 2 5 8 3 6 9];  %Block interleaver, N=9
Inter=[1 5 9 13 2 6 10 14 3 7 11 15 4 8 12 16];  %Block interleaver, N=16
%Inter=[1 6 11 16 21 2 7 12 17 22 3 8 13 18 23 4 9 14 19 24 5 10 15 20 25];
%%Block interleaver, N=25
gmax=1;
 
for g=1:gmax;    % This loop runs the encoding of a block gmax times   
   
	InputBits = rand(N-2,1);    % A randomly generated input message
	for i= 1:N-2
   	if InputBits(i)>0.5 
      	InputBits(i)=1;         % polar format for input message
   	else
      	InputBits(i)=-1;
   	end
	end

	TrellisInput(1)=1;         % Trellis coded sequence corresponding to the input message block,until N-2
	for i= 1:N-2;              % Last two bits are then determined to make the coded sequence be terminated
		for j= 1:4;
   		if Mat(TrellisInput(i),j,1)==InputBits(i);
         	TrellisInput(i+1)=j;
         	Output(i,1)= Mat(TrellisInput(i),j,2);
         	Output(i,2)= Mat(TrellisInput(i),j,3);
      	end
   	end
	end

	switch TrellisInput(N-1)  % This rutine forces the last two bits to make the coded sequence be terminated 
	case 1                                              
   	InputBits(N-1)=-1;InputBits(N)=-1;
	case 2
   	InputBits(N-1)=-1;InputBits(N)=+1;
	case 3
   	InputBits(N-1)=+1;InputBits(N)=-1;
	case 4
   	InputBits(N-1)=+1;InputBits(N)=+1;
	end

	for i= N-1:N;            % This rutine completes the coded sequence with the last tow message bits
		for j= 1:4;
   		if Mat(TrellisInput(i),j,1)==InputBits(i);
         	TrellisInput(i+1)=j;
         	Output(i,1)= Mat(TrellisInput(i),j,2);
         	Output(i,2)= Mat(TrellisInput(i),j,3);
      	end
   	end
	end	
%The variable Output is the output of the turbo code, a sequence to be
%transmitted
	
   	InputBits2(:,1)= InputBits(Inter(:),1);    % Interleaving of the input message to be input 
                                                % the second encoder
   
   TrellisInput2(1)=1; %generation of the encoded sequence of the second encoder
	for i= 1:N;
		for j= 1:4;
   		if Mat(TrellisInput2(i),j,1)==InputBits2(i);
         	TrellisInput2(i+1)=j;
         	Output2(i,1)= Mat(TrellisInput2(i),j,2);
         	Output2(i,2)= Mat(TrellisInput2(i),j,3);
      	end
   	end
	end
   

	for i= 2:2:N;                   % Bits at even positions of the second encoder are placed 
   	Output(i,2)=Output2(i,2);       % at the variable output, the bits to be transmitted
   end
 	%This rutine ends the puncturing procedure
    
	
    

	Sigma = 1.2; %Transmited bits are affected by AWGN, function randn performs this step
	Noise = Sigma*randn(N,2);
	%ReceivedBits = Output;           % noise free case
	ReceivedBits = Output+Noise;        
	Lc = 2/(Sigma^2);                % Lc=4.Eb/(2.Sigma^2),   with  Eb=1

	% In the following the received sequences for each decoder, Y1 and Y2, are formed
    % tacking into account the interleaving and puncturing procedures
   
   Y1=zeros(N,2);
	Y2=Y1;
	Y1=ReceivedBits;
	for i= 2:2:N
   	Y1(i,2)=0;   % A zero is placed at the even positions for the first decoder, 
    end            %this takes into account the puncturing

	for i= 1:N;
   	Y2(i,1)= ReceivedBits(Inter(i),1);       %Received sequence is interleaved for the second decoder
   end
      
	for i=2:2:N
   	Y2(i,2)=ReceivedBits(i,2);
	end
   	
	%The following is the logmap BCJR decoder operation
    
	Luk = zeros(N,1); %initialization of the a priori infomation Luk
	Gama= zeros(N,4,4);%initialization of coefficients gamma

	for h=1:7;      %iterative decoding Loop
                    %Each loop is related to the operation of one of the decoders
      
      if mod(h,2)==1;%decision about the decoder that has to operate
      	Y=Y1;
   	else
      	Y=Y2; 
   	end
      	 
   	for i= 1:N;  %Coefficients gamma are here calculated
      	for k= 1:4;
         	for j= 1:4;
            	if Mat(j,k,1)~=0;
               	Gama(i,j,k)= exp(Mat(j,k,1).*Luk(i)/2).*exp(Lc/2*(Y(i,1).*Mat(j,k,2)+Y(i,2).*Mat(j,k,3)));
            	end
         	end
      	end
      end
            	   
   	Alfa=zeros(N,4);  % %Coefficients alfa are here calculated
   	Alfa(1,1)=1;   
   	for i=2:N;
      	for k=1:4;
         	for j=1:4;
            	Alfa(i,k)=Alfa(i,k)+(Gama(i-1,j,k).*Alfa(i-1,j));
         	end
      	end
   	end
      	   
   	Beta=zeros(N+1,4);   % %Coefficients beta are here calculated, they are all first set to 0
   	if mod(h,2)==1;
      	Beta(N+1,1)=1;    %For the first decoder, Beta(N+1,1) should be 1 y and he others should be 0
   	else
      	Beta(N+1,:)=1;    %For the second decoder all the beta values are 1
      end                  % because encoder 2 produces a non terminated sequence   
   	for i=N:-1:2;
      	for k=1:4;
         	for j=1:4;
            	Beta(i,k)=Beta(i,k)+(Beta(i+1,j).*Gama(i,k,j));
         	end
      	end
   	end
      
   	Luky=zeros(N,1);   % Luky is the conditional estimation L(b(i)/y)
   	for i=1:N;
      	Numer=0;
      	Denom=0;
      	for k=1:4;
         	for j=1:4;
            	if Mat(j,k,1)==1;
               	Numer = Numer + (Alfa(i,j).*Gama(i,j,k).*Beta(i+1,k));
            	elseif Mat(j,k,1)==-1;
               	Denom = Denom + (Alfa(i,j).*Gama(i,j,k).*Beta(i+1,k));
            	end
         	end      
      	end
      	Luky(i)=log(Numer/Denom); %Estimations of the bits
   	end
     	LukyHard=sign(Luky);    %Hard decision at the coresponding iteration 
                                %is the sign of the bit estimation
        
             
   	% Calculation of the extrinsic information to be passed to the other
   	% decoder
   
   	Le(:,1) = Luky(:,1) - Lc.*Y(:,1) - Luk(:,1);
   
   	
         Luk(i)= Le(Inter(i));  % De-interleaving of the extrinsic information 
                                % that is the a priori information passed to the
                                %other decoder
       
	   
end      %End of the iterative decoding loop 

Errores= Errores + sum(abs(LukyHard-InputBits)/2);
   
end         % End of the loop of variable g 

Errores